//********************************************************************
UNIT Math3D;
//********************************************************************
INTERFACE

USES
  Classes, Sysutils, StdBasicDefs;

TYPE

   TVector3D = Record X,Y,Z: TFloat32; End;
   TVector4D = Record X,Y,Z,W: TFloat32; End;
   TMatrix3D = Array[0..2] Of TVector3D;
   TMatrix4D = Array[0..3] Of TVector4D;
   PMatrix4D = ^TMatrix4D;

   TRay = Record Point,DirVector: TVector4D; End;
   TPlane = Record Point,Normal: TVector4D; End;
   TSphere = Record Center: TVector4D; Radius: TFloat32; End;

CONST
   MATIDENTITY: TMatrix4D = (( X:1; Y:0; Z:0; W:0),
                             ( X:0; Y:1; Z:0; W:0),
                             ( X:0; Y:0; Z:1; W:0),
                             ( X:0; Y:0; Z:0; W:1));

   ZEROVECTOR: TVector4D =   ( X:0; Y:0; Z:0; W:1);

   PiDiv180 = PI/180;

//********************************************************************
Function MatrixMultiply(M1,M2: TMatrix4D): TMatrix4D;
Function MatrixRotate(Angle,AX,AY,AZ: TFloat32): TMatrix4D;
Function MatrixShift(AX,AY,AZ: TFloat32): TMatrix4D;
//********************************************************************
Function Vector3DTo4D(AVector: TVector3D): TVector4D;
Function Vector4DTo3D(AVector: TVector4D): TVector3D;
Function VectorAdd(Vector1,Vector2: TVector4D): TVector4D;
Function VectorCrossProduct(Vector1,Vector2: TVector4D): TVector4D;
Function VectorDeduct(FromPoint,ToPoint: TVector4D): TVector4D;
Function VectorDotProduct(AVector1, AVector2: TVector4D): TFloat32;
Function VectorInvers(AVector: TVector4D): TVector4D;
Function VectorMultiply(AVector: TVector4D;
                   AFactor: TFloat32): TVector4D;
Function VectorTransform(AMatrix: TMatrix4D;
                   AVector: TVector4D): TVector4D;
Function VectorValue(AVector: TVector4D): TFloat32;
//********************************************************************
Function Det3x3(AMatrix3D: TMatrix3D): TFloat32;
Function LinearResult(AMatrix3D: TMatrix3D;
                   ARightSide: TVector3D): TVector3D;
//********************************************************************
IMPLEMENTATION
//********************************************************************
Function MatrixMultiply(M1,M2: TMatrix4D): TMatrix4D;
Var Col: TInt32;
Begin
   For Col:= 0 To 3 Do With Result[Col] Do Begin
      X:= M1[0].X * M2[Col].X +
          M1[1].X * M2[Col].Y +
          M1[2].X * M2[Col].Z +
          M1[3].X * M2[Col].W;

      Y:= M1[0].Y * M2[Col].X +
          M1[1].Y * M2[Col].Y +
          M1[2].Y * M2[Col].Z +
          M1[3].Y * M2[Col].W;

      Z:= M1[0].Z * M2[Col].X +
          M1[1].Z * M2[Col].Y +
          M1[2].Z * M2[Col].Z +
          M1[3].Z * M2[Col].W;

      W:= M1[0].W * M2[Col].X +
          M1[1].W * M2[Col].Y +
          M1[2].W * M2[Col].Z +
          M1[3].W * M2[Col].W;
   End;
End;
//********************************************************************
Function MatrixRotate(Angle,AX,AY,AZ: TFloat32): TMatrix4D;
Var
   Sinus,Cosinus: TFloat32;
Begin
   Cosinus:= Cos(Angle*PiDiv180);
   Sinus:= Sin(Angle*PiDiv180);
   //Sinus:= Sqrt(1-Cosinus*Cosinus); The sign is lost here!

   Result:= MATIDENTITY;

   Result[0].X:= AX*AX*(1-Cosinus) + Cosinus;
   Result[0].Y:= AX*AY*(1-Cosinus) + AZ*Sinus;
   Result[0].Z:= AX*AZ*(1-Cosinus) - AY*Sinus;

   Result[1].X:= AY*AX*(1-Cosinus) - AZ*Sinus;
   Result[1].Y:= AY*AY*(1-Cosinus) + Cosinus;
   Result[1].Z:= AY*AZ*(1-Cosinus) + AX*Sinus;

   Result[2].X:= AZ*AX*(1-Cosinus) + AY*Sinus;
   Result[2].Y:= AZ*AY*(1-Cosinus) - AX*Sinus;
   Result[2].Z:= AZ*AZ*(1-Cosinus) + Cosinus;
End;
//********************************************************************
Function MatrixShift(AX,AY,AZ: TFloat32): TMatrix4D;
Begin
   Result:= MATIDENTITY;
   With Result[3] Do Begin
      X:= AX;
      Y:= AY;
      Z:= AZ;
   End;
End;
//********************************************************************
Function Vector3DTo4D(AVector: TVector3D): TVector4D;
Begin
   With Result Do Begin
      X:= AVector.X;
      Y:= AVector.Y;
      Z:= AVector.Z;
      W:= 1;
   End;
End;
//********************************************************************
Function Vector4DTo3D(AVector: TVector4D): TVector3D;
Begin
   With Result Do Begin
      X:= AVector.X;
      Y:= AVector.Y;
      Z:= AVector.Z;
   End;
End;
//********************************************************************
Function VectorAdd(Vector1,Vector2: TVector4D): TVector4D;
Begin
   Result.X:= Vector1.X + Vector2.X;
   Result.Y:= Vector1.Y + Vector2.Y;
   Result.Z:= Vector1.Z + Vector2.Z;
   Result.W:= 1;
End;
//********************************************************************
Function VectorCrossProduct(Vector1,Vector2: TVector4D): TVector4D;
Begin
   Result.X:= Vector1.Y*Vector2.Z - Vector1.Z*Vector2.Y;
   Result.Y:= Vector1.Z*Vector2.X - Vector1.X*Vector2.Z;
   Result.Z:= Vector1.X*Vector2.Y - Vector1.Y*Vector2.X;
   Result.W:= 1;
End;
//********************************************************************
Function VectorDeduct(FromPoint,ToPoint: TVector4D): TVector4D;
Begin
   Result.X:= ToPoint.X - FromPoint.X;
   Result.Y:= ToPoint.Y - FromPoint.Y;
   Result.Z:= ToPoint.Z - FromPoint.Z;
   Result.W:= 1;
End;
//********************************************************************
Function VectorDotProduct(AVector1, AVector2: TVector4D): TFloat32;
Begin
   Result:= AVector1.X * AVector2.X +
            AVector1.Y * AVector2.Y +
            AVector1.Z * AVector2.Z;
End;
//********************************************************************
Function VectorInvers(AVector: TVector4D): TVector4D;
Begin
   With Result Do Begin
      X:= -AVector.X;
      Y:= -AVector.Y;
      Z:= -AVector.Z;
      W:= 1;
   End;
End;
//********************************************************************
Function VectorMultiply(AVector: TVector4D;
                        AFactor: TFloat32): TVector4D;
Begin
   With Result Do Begin
      X:= AFactor * AVector.X;
      Y:= AFactor * AVector.Y;
      Z:= AFactor * AVector.Z;
      W:= 1;
   End;
End;
//********************************************************************
Function VectorTransform(AMatrix: TMatrix4D;
                         AVector: TVector4D): TVector4D;
Begin
   With AVector Do Begin
      Result.X:= X * AMatrix[0].X +
                 Y * AMatrix[1].X +
                 Z * AMatrix[2].X +
                 W * AMatrix[3].X;

      Result.Y:= X * AMatrix[0].Y +
                 Y * AMatrix[1].Y +
                 Z * AMatrix[2].Y +
                 W * AMatrix[3].Y;

      Result.Z:= X * AMatrix[0].Z +
                 Y * AMatrix[1].Z +
                 Z * AMatrix[2].Z +
                 W * AMatrix[3].Z;

      Result.W:= X * AMatrix[0].W +
                 Y * AMatrix[1].W +
                 Z * AMatrix[2].W +
                 W * AMatrix[3].W;
   End;

   With Result Do If Abs(W) > Tolerance
      Then Begin
         X:=X/W;
         Y:=Y/W;
         Z:=Z/W;
         W:=1;
      End;
End;
//********************************************************************
Function VectorValue(AVector: TVector4D): TFloat32;
Begin With AVector Do Result:= Sqrt(X*X+Y*Y+Z*Z); End;
//********************************************************************
Function Det3x3(AMatrix3D: TMatrix3D): TFloat32;
Begin
   Result:= + AMatrix3D[0].X * AMatrix3D[1].Y * AMatrix3D[2].Z
            + AMatrix3D[0].Y * AMatrix3D[1].Z * AMatrix3D[2].X
            + AMatrix3D[0].Z * AMatrix3D[1].X * AMatrix3D[2].Y

            - AMatrix3D[2].X * AMatrix3D[1].Y * AMatrix3D[0].Z
            - AMatrix3D[2].Y * AMatrix3D[1].Z * AMatrix3D[0].X
            - AMatrix3D[2].Z * AMatrix3D[1].X * AMatrix3D[0].Y;
End;
//********************************************************************
Function LinearResult(AMatrix3D: TMatrix3D;
                   ARightSide: TVector3D): TVector3D;

Var TempMat: TMatrix3D; BaseDet,XDet,YDet,ZDet: TFloat32;
Begin
   BaseDet:= Det3x3(AMatrix3D);

   TempMat:= AMatrix3D;
   TempMat[0]:= ARightSide;
   XDet:= Det3x3(TempMat);

   TempMat:= AMatrix3D;
   TempMat[1]:= ARightSide;
   YDet:= Det3x3(TempMat);

   TempMat:= AMatrix3D;
   TempMat[2]:= ARightSide;
   ZDet:= Det3x3(TempMat);

   Result.X:= XDet/BaseDet;
   Result.Y:= YDet/BaseDet;
   Result.Z:= ZDet/BaseDet;
End;
//********************************************************************
//********************************************************************
End.

