//********************************************************************
UNIT Shape3D;
//********************************************************************
INTERFACE

USES
   SysUtils, Classes, Contnrs, Math,
   StdBasicDefs, Math3D, OSSupport, GraphicSupport;

TYPE

   TShape3DType = (st3Triangle, st3Rectangle ,st3Cube, st3Sphere);

   TAnimation = Record
      StartMatrix: TMatrix4D;
      MilliSecs: TuInt32;
      Speed: TFloat32;
   End;

//********************************************************************
TYPE

   TShape3D = Class(TObject)
   Private
      // Visual properties
      fShape3DType: TShape3DType;
      fVisible: TBool8;
      fScale: TFloat32;
      fColor: T32bColor;
      fSaveColor: T32bColor;
      fColCount,fRowCount: TInt32;

      // Animation properties
      fAnimation: TAnimation;
      fMatrix: TMatrix4D;
      fAngle: TFloat32;
      fStartTime,fStopTime: TInt64;
      fBdgSphere: TSphere;

      // Render buffers
      fVertexBuffer: TVertexBuffer;
      fNormalBuffer: TVertexBuffer;
      fIndexBuffer: TIndexBuffer;

      Procedure CreateTriangle;
      Procedure CreateRectangle;
      Procedure CreateCube;
      Procedure CreateSphere;

       Function GetMatrix: TMatrix4D;
      Procedure SetMatrix(AMatrix: TMatrix4D);

      Procedure CreateBoundingSphere;
       Function CreateSphereVertex(ACol,ARow: Cardinal): TVector3D;
      Procedure SetVertex(AIndex: TuInt32; AX,AY,AZ: TFloat32);
      Procedure SetNormal(AIndex: TuInt32; AX,AY,AZ: TFloat32);
      Procedure SetRectangleBuffer(AIndex,AVertex: TuInt32);

   Public
      Property Matrix: TMatrix4D Read GetMatrix Write SetMatrix;
      Property ShapeType: TShape3DType Read fShape3DType;

      Constructor Create(AShapeType: TShape3DType; AScale: TFloat32;
                   AColor: T32bColor); Reintroduce;
       Destructor Destroy; Override;

         Function Hit(Opponent: TShape3D): TBool8;
         Function HitEdge(Opponent: TShape3D): TBool8;

        Procedure Animate(AStartMatrix: TMatrix4D;
                   AMilliSecs: TuInt32; ASpeed: TFloat32);
        Procedure StartAnimation;
        Procedure StopAnimation;
        Procedure RunAnimation;
        Procedure DefaultAnimation;
   End;

//********************************************************************
   TShapeMan = Class(TObject)
   Private
      fObjectList: TObjectList;
      fViewMatrix: TMatrix4D;

      Procedure CollisionTest;

   Public
      Constructor Create; Reintroduce;
       Destructor Destroy; Override;
        Procedure Execute;

        Function CreateShape3D(AShapeType: TShape3DType;
                   AScale: TFloat32; AColor: T32bColor): TShape3D;
   End;

VAR
   ShapeMan: TShapeMan;

//********************************************************************
IMPLEMENTATION
//********************************************************************
//********************************************************************
//   T S H A P E M A N
//********************************************************************
Constructor TShapeMan.Create;
Const
   ZNEAR = 1;
   ZFAR = 30;
Begin
   Inherited Create;
   fObjectList:= TObjectList.Create;
   fViewMatrix:= MatrixMultiPly(MatrixRotate(+30,1,0,0),
                                MatrixRotate(-30,0,1,0));
   fViewMatrix:= MatrixMultiPly(MatrixShift(0,0,-12),fViewMatrix);
   SetViewMatrix(fViewMatrix);

   InitGraphicParams(ZNEAR,ZFAR);
   CreateGraphicTools;
End;
//********************************************************************
Destructor TShapeMan.Destroy;
Begin
   FreeGraphicTools;
   fObjectList.Free;
   Inherited Destroy;
End;
//********************************************************************
Procedure TShapeMan.Execute;
Var LocalShape: TShape3D; I: TuInt32;
Begin
   With fObjectList Do If Count > 0 Then Begin

      BeginRender;

      For I:= 0 To Count-1 Do Begin
          LocalShape:= TShape3D(Items[I]);
          With LocalShape Do If fVisible Then Begin

             RunAnimation;
             CollisionTest;
             Render(fMatrix,fColor,
                fVertexBuffer,fNormalBuffer,fIndexBuffer);

          End;
      End;

      EndRender;
   End;
End;
//********************************************************************
Function TShapeMan.CreateShape3D(AShapeType: TShape3DType;
                   AScale: TFloat32; AColor: T32bColor): TShape3D;
Var LocalShape: TShape3D;
Begin
   LocalShape:= TShape3D.Create(AShapeType,AScale,AColor);
   fObjectList.Add(LocalShape);
   Result:= LocalShape;
End;
//********************************************************************
Procedure TShapeMan.CollisionTest;
Var LocalShape,Opponent: TShape3D; I,J: TuInt32;
Begin
   With fObjectList Do If Count > 0
      Then For I:= 0 To Count-1 Do Begin
          LocalShape:= TShape3D(Items[I]);
          For J:= 0 To Count-1 Do Begin
             Opponent:= TShape3D(Items[J]);
                If LocalShape <> Opponent
                   Then If LocalShape.Hit(Opponent)
                      Then;
          End;
      End;
End;
//********************************************************************
//********************************************************************
//   T S H A P E 3 D
//********************************************************************
Constructor TShape3D.Create(AShapeType: TShape3DType;
                   AScale: TFloat32; AColor: T32bColor);
Begin
   fShape3DType:= AShapeType;
   fScale:= AScale;
   fColor:= AColor;
   fSaveColor:= AColor;

   fVisible:= TRUE;
   Matrix:= MATIDENTITY;
   fColCount:= 18;
   fRowCount:= 9;

   fAngle:= 0;

   With fAnimation Do Begin
      StartMatrix:= MATIDENTITY;;
      MilliSecs:= 0;
      Speed:= 0;
   End;

   Case fShape3DType Of

       st3Triangle: CreateTriangle;
      st3Rectangle: CreateRectangle;
           st3Cube: CreateCube;
         st3Sphere: CreateSphere;

   End;//Of Case

   CreateBoundingSphere;
End;
//********************************************************************
Destructor TShape3D.Destroy;
Begin
   Finalize(fIndexBuffer);
   Finalize(fNormalBuffer);
   Finalize(fVertexBuffer);
End;
//********************************************************************
Function TShape3D.Hit(Opponent: TShape3D): TBool8;
Var
   ActDistance,MinDistance: TFloat32;
Begin
   fColor:= fSaveColor;

   ActDistance:= VectorValue(VectorDeduct
                 (fBdgSphere.Center,Opponent.fBdgSphere.Center));

   MinDistance:=  fBdgSphere.Radius+Opponent.fBdgSphere.Radius;

   If ActDistance < MinDistance
      Then Begin

         If (Opponent.fShape3DType = st3Triangle) And
            (HitEdge(Opponent))
               Then Begin
                  fColor:= SRED;
                  Result:= TRUE;
               End;
      End
      Else Result:= FALSE;
End;
//********************************************************************
Function TShape3D.HitEdge(Opponent: TShape3D): TBool8;
Var
   HitPoint: TVector4D;
   TriangleP0,
   TriangleP1: TVector4D; // Transformed Triangle Points
   Edge: TVector4D;       // Polygonedge
   BtV: TVector4D;        // Sphere Center => PolygonVertex
   Vel: TVector4D;        // SphereDir * speed
   SqEdge: TFloat32;      // SquareLength of edge
   SqBtV: TFloat32;       // SquareLength of BtV
   SqVel: TFloat32;       // SquareLength of Velocity
   DotEdgeVel: TFloat32;
   DotEdgeBtV: TFloat32;
   DotVelBtV: TFloat32;
   Distance: TFloat32;
   a,b,c,Discr,SqrResult,f: TFloat32;
Begin
   Result:= FALSE;

   With Opponent Do Begin
      TriangleP0:= VectorTransform
         (fMatrix,Vector3DTo4D(fVertexBuffer[0]));
      TriangleP1:= VectorTransform
         (fMatrix,Vector3DTo4D(fVertexBuffer[1]));
   End;

   Edge:= VectorDeduct(TriangleP1,TriangleP0);
   BtV := VectorDeduct(fBdgSphere.Center,TriangleP0);
   Vel := VectorMultiply(fMatrix[2],fAnimation.Speed);

   SqEdge:= Sqr(VectorValue(Edge));
   SqBtV := Sqr(VectorValue(BtV));
   SqVel := Sqr(VectorValue(Vel));

   DotEdgeVel:= VectorDotProduct(Edge,Vel);
   DotEdgeBtV:= VectorDotProduct(Edge,BtV);
   DotVelBtV := VectorDotProduct(Vel,BtV);

   a:= SqEdge*(-SqVel) + DotEdgeVel*DotEdgeVel;
   b:= 2*(sqEdge*DotVelBtV - DotEdgeVel*DotEdgeBtV);
   With fBdgSphere Do
      c:= sqEdge*(Radius*Radius-SqBtV)+DotEdgeBtV*DotEdgeBtV;

   Discr:= b*b - 4*a*c;

   If Discr >= 0 Then Begin
      SqrResult:= (-b + Sqrt(Discr)) / (2*a);

      If SqrResult > 0 Then Begin
         f:= (DotEdgeVel*SqrResult - DotEdgeBtV) / SqEdge;
         Distance:= SqrResult*VectorValue(Vel);
         //HitPoint:= VectorAdd(TriangleP1,VectorMultiply(Edge,f));
         If Distance < 0.05
            Then Result:= TRUE;
      End;
   End;
End;
//********************************************************************
Procedure TShape3D.Animate(AStartMatrix: TMatrix4D;
                   AMilliSecs: TuInt32; ASpeed: TFloat32);
Begin
   With fAnimation Do Begin
      StartMatrix:= AStartMatrix;
      MilliSecs:= AMilliSecs;
      Speed:= ASpeed;
   End;
   StartAnimation;
End;
//********************************************************************
Procedure TShape3D.StartAnimation;
Begin
   Matrix:= fAnimation.StartMatrix;
   fStartTime:= osClockTick;
   fStopTime:= fStartTime+fAnimation.MilliSecs;
End;
//********************************************************************
Procedure TShape3D.StopAnimation;
Begin
   Matrix:= MATIDENTITY;
   fStartTime:= 0;
   fStopTime:= 0;
End;
//********************************************************************
Procedure TShape3D.RunAnimation;
Var Shift: TVector4D; ActTime: TInt64;
Begin
   If fStartTime <> 0
      Then Begin
         ActTime:= osClockTick;
         If ActTime < fStopTime
            Then Begin
               Shift:= VectorMultiply(fMatrix[2],fAnimation.Speed);
               fMatrix[3]:= VectorAdd(fMatrix[3],Shift);
               Matrix:= fMatrix;
            End
            Else StartAnimation;
      End;
      //Else DefaultAnimation;
End;
//********************************************************************
Procedure TShape3D.DefaultAnimation;
Var Mat1,Mat2: TMatrix4D;
Begin
   fAngle:= fAngle+1;
   If fAngle >= 360
      Then fAngle:= 0;

   Matrix:= MatrixRotate(fAngle,0,1,0);
End;
//********************************************************************
//********************************************************************
Procedure TShape3D.CreateTriangle;
Var Sin30,Cos30: TFloat32; I: TuInt8;
Begin
   SetLength(fVertexBuffer,3);
   SetLength(fNormalBuffer,3);
   SetLength(fIndexBuffer,3);

   Sin30:= Sin(30*PiDiv180);
   Cos30:= Cos(30*PiDiv180);

   SetVertex(0, 0,1,0);
   SetVertex(1, -Cos30,-Sin30,0);
   SetVertex(2, +Cos30,-Sin30,0);

   For I:= 0 To 2 Do SetNormal(I,0,0,1);

   For I:= 0 To 2 Do fIndexBuffer[I]:= I;
End;
//********************************************************************
Procedure TShape3D.CreateRectangle;
Var I: TuInt8;
Begin
   SetLength(fVertexBuffer,4);
   SetLength(fNormalBuffer,4);
   SetLength(fIndexBuffer,6);

   SetVertex(0, -1,-1,0);
   SetVertex(1, +1,-1,0);
   SetVertex(2, +1,+1,0);
   SetVertex(3, -1,+1,0);

   For I:= 0 To 3 Do SetNormal(I,0,0,1);
   SetRectangleBuffer(0,0);
End;
//********************************************************************
Procedure TShape3D.CreateCube;
Var I: TuInt8;
Begin
   SetLength(fVertexBuffer,24);
   SetLength(fNormalBuffer,24);
   SetLength(fIndexBuffer,36);

   // Front
   SetVertex(0, -1,-1,+1);
   SetVertex(1, +1,-1,+1);
   SetVertex(2, +1,+1,+1);
   SetVertex(3, -1,+1,+1);

   For I:= 0 To 3 Do SetNormal(I,+0,+0,+1);
   SetRectangleBuffer(0,0);

   // Back
   SetVertex(4, +1,-1,-1);
   SetVertex(5, -1,-1,-1);
   SetVertex(6, -1,+1,-1);
   SetVertex(7, +1,+1,-1);

   For I:= 4 To 7 Do SetNormal(I,+0,+0,-1);
   SetRectangleBuffer(6,4);

   // Left
   SetVertex( 8, -1,-1,-1);
   SetVertex( 9, -1,-1,+1);
   SetVertex(10, -1,+1,+1);
   SetVertex(11, -1,+1,-1);

   For I:= 8 To 11 Do SetNormal(I,-1,+0,+0);
   SetRectangleBuffer(12,8);

   // Right
   SetVertex(12, +1,-1,+1);
   SetVertex(13, +1,-1,-1);
   SetVertex(14, +1,+1,-1);
   SetVertex(15, +1,+1,+1);

   For I:= 12 To 15 Do SetNormal(I,+1,+0,+0);
   SetRectangleBuffer(18,12);

   // Top
   SetVertex(16, -1,+1,+1);
   SetVertex(17, +1,+1,+1);
   SetVertex(18, +1,+1,-1);
   SetVertex(19, -1,+1,-1);

   For I:= 16 To 19 Do SetNormal(I,+0,+1,+0);
   SetRectangleBuffer(24,16);

   // Bottom
   SetVertex(20, -1,-1,-1);
   SetVertex(21, +1,-1,-1);
   SetVertex(22, +1,-1,+1);
   SetVertex(23, -1,-1,+1);

   For I:= 20 To 23 Do SetNormal(I,+0,-1,+0);
   SetRectangleBuffer(30,20);
End;
//********************************************************************
Procedure TShape3D.CreateSphere;
Var VertexCount,CellCount,IndexCount,
    VtxIndex,IdxIndex,Col,Row: TInt32;
Begin
   VertexCount:= (fColCount+1)*(fRowCount+1);
   CellCount:= fColCount*fRowCount;
   IndexCount:= CellCount*6;

   SetLength(fVertexBuffer,VertexCount);
   SetLength(fNormalBuffer,VertexCount);
   SetLength(fIndexBuffer,IndexCount);

   // Create Vertices and Normals
   VtxIndex:= 0;
   For Row:= 0 To fRowCount Do
      For Col:= 0 To fColCount Do Begin
         With CreateSphereVertex(Col,Row) Do
            SetVertex(VtxIndex,X,Y,Z);
         fNormalBuffer[VtxIndex]:= fVertexBuffer[VtxIndex];
         Inc(VtxIndex);
      End;

   // Create Indices
   VtxIndex:= 0;
   IdxIndex:= 0;

   For Row:= 0 To fRowCount-1 Do Begin
      For Col:= 0 To fColCount-1 Do Begin

         fIndexBuffer[IdxIndex+0]:= VtxIndex+fColCount+2;
         fIndexBuffer[IdxIndex+1]:= VtxIndex+fColCount+1;
         fIndexBuffer[IdxIndex+2]:= VtxIndex+1;

         fIndexBuffer[IdxIndex+3]:= VtxIndex+fColCount+1;
         fIndexBuffer[IdxIndex+4]:= VtxIndex;
         fIndexBuffer[IdxIndex+5]:= VtxIndex+1;

         Inc(VtxIndex);
         Inc(IdxIndex,6);
      End;
      Inc(VtxIndex);
   End;
End;
//********************************************************************
//********************************************************************
Function TShape3D.GetMatrix: TMatrix4D;
Begin
   Result:= fMatrix;
End;
//********************************************************************
Procedure TShape3D.SetMatrix(AMatrix: TMatrix4D);
Begin
   fMatrix:= AMatrix;
   fBdgSphere.Center:= fMatrix[3];
End;
//********************************************************************
Procedure TShape3D.CreateBoundingSphere;
Var VectorLen: TFloat32; I: TuInt32;
Begin
   With fBdgSphere Do Begin
      Center:= ZEROVECTOR;
      Radius:= 0;
   End;

   For I:= 0 To High(fVertexBuffer) Do Begin
      VectorLen:= VectorValue(Vector3DTo4D(fVertexBuffer[I]));
      With fBdgSphere Do Radius:= Max(Radius,VectorLen);
   End;
End;
//********************************************************************
Function TShape3D.CreateSphereVertex(ACol,ARow: Cardinal): TVector3D;
Var VecSize: TFloat32;
Begin
   With Result Do Begin
      X:= Sin(180 Div fRowCount*ARow*PiDiv180)
                     * Cos(360 Div fColCount*ACol*PiDiv180);

      Y:= Cos(180 Div fRowCount*ARow*PiDiv180);

      Z:= Sin(180 Div fRowCount*ARow*PiDiv180)
                     * Sin(360 Div fColCount*ACol*PiDiv180);
   End;
End;
//********************************************************************
Procedure TShape3D.SetVertex(AIndex: TuInt32; AX,AY,AZ: TFloat32);
Begin
   With fVertexBuffer[AIndex]
      Do Begin
         X:=fScale*AX;
         Y:=fScale*AY;
         Z:=fScale*AZ;
      End;
End;
//********************************************************************
Procedure TShape3D.SetNormal(AIndex: TuInt32; AX,AY,AZ: TFloat32);
Begin
   With fNormalBuffer[AIndex] Do Begin X:=AX; Y:=AY; Z:=AZ; End;
End;
//********************************************************************
Procedure TShape3D.SetRectangleBuffer(AIndex,AVertex: TuInt32);
Begin
   fIndexBuffer[AIndex+0]:= AVertex+0;
   fIndexBuffer[AIndex+1]:= AVertex+1;
   fIndexBuffer[AIndex+2]:= AVertex+3;
   fIndexBuffer[AIndex+3]:= AVertex+1;
   fIndexBuffer[AIndex+4]:= AVertex+2;
   fIndexBuffer[AIndex+5]:= AVertex+3;
End;
//********************************************************************
//********************************************************************
End.

